/*
    This file is part of ben -- a task scheduler for distributed computations
    Copyright (C) 2012-2019  Laurent Poirrier

    ben is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <stdarg.h>
#include <errno.h>
#include <string.h>
#include "xmem.h"
#include "config.h"
#include "buffer.h"

#define DEBUG(...)	fprintf(stderr, __VA_ARGS__)

// **************************************************************************
// 
// **************************************************************************
void buffer_init(buffer *b, size_t gran)
{
	size_t m;

	if (gran == 0) {
		b->mask = 0;
	} else {
		m = 2;
	
		while ((m < gran) && (m))
			m <<= 1;
	
		if (m)
			b->mask = m - 1;
		else
			b->mask = 0;
	}

	b->size = 0;
	b->alloc = 0;
	b->partial = 0;
	b->data = NULL;
	b->eof = 0;
	b->error = 0;
	b->total_read = 0;
	b->total_write = 0;
}

// **************************************************************************
// 
// **************************************************************************
void buffer_clear(buffer *b)
{
	xfree(b->data);
}

// **************************************************************************
// 
// **************************************************************************
void buffer_reset(buffer *b)
{
	b->size = 0;
}

// **************************************************************************
// 
// **************************************************************************
void buffer_alloc(buffer *b, size_t size)
{
	if (size <= b->alloc)
		return;

	if (b->mask) {
		b->alloc = (size + b->mask) & (~b->mask);
	} else {
		b->alloc = 1;
		
		while ((b->alloc < size) && (b->alloc))
			b->alloc <<= 1;
		
		if (!b->alloc) {
			fprintf(stderr, "buffer_alloc(): %zd too large\n",
				size);
			exit(1);
		}
	}
	
	b->data = xrealloc(b->data, b->alloc);
}

void *buffer_get(buffer *b, size_t size)
{
	buffer_alloc(b, size);
	b->size = size;
	return(b->data);
}

// **************************************************************************
// 
// **************************************************************************
void buffer_memcpy(buffer *b, const void *src, size_t size)
{
	buffer_get(b, size);
	memcpy(b->data, src, size);
}

void buffer_memappend(buffer *b, const void *src, size_t size)
{
	size_t o;
	
	o = b->size;
	buffer_get(b, o + size);
	
	memcpy(b->data + o, src, size);
}

// **************************************************************************
// 
// **************************************************************************
char *buffer_str(buffer *b)
{
	buffer_get(b, b->size + 1);
	b->size--;
	((char *)b->data)[b->size] = 0;
	return((char *)b->data);
}

void buffer_strcpy(buffer *b, const char *src)
{
	size_t len;
	
	len = strlen(src);
	buffer_memcpy(b, src, len);
}

void buffer_strappend(buffer *b, const char *src)
{
	size_t len;
	
	len = strlen(src);
	buffer_memappend(b, src, len);
}

// **************************************************************************
// 
// **************************************************************************
char *buffer_fappend(buffer *b, const char *format, ...)
{
	size_t o;
	int len;
	va_list ap;
	
	va_start(ap, format);
	len = vsnprintf(NULL, 0, format, ap);
	va_end(ap);

	o = b->size;
	buffer_alloc(b, o + len + 1);
	buffer_get(b, o + len);

	va_start(ap, format);
	vsnprintf(b->data + o, len + 1, format, ap);
	va_end(ap);

	return((char *)b->data + o);
}


// **************************************************************************
// 
// **************************************************************************
int buffer_load(buffer *b, int fd, size_t max_size)
{
	size_t o, block;
	ssize_t n;
	
	if (b->mask)
		block = b->mask + 1;
	else
		block = conf.block_size;
	
	o = 0;
	while (1) {
		buffer_alloc(b, o + block);
		
		n = read(fd, b->data + o, b->alloc - o);
		
		if (n < 0) {
			if (errno == EINTR) {
				errno = 0;
			} else {
				fprintf(stderr, "buffer_load(): read(): %s\n",
					strerror(errno));
				return(-1);
			}
		} else if (n == 0) {
			b->size = o;
			return(0);
		} else {
			o += n;
			b->total_read += n;
		}

		if (o > max_size) {
			fprintf(stderr, "buffer_load(): size cap hit (%zd)\n",
				max_size);
			return(-1);
		}
	}
}

int buffer_save(buffer *b, int fd)
{
	size_t o;
	ssize_t n;
	
	o = 0;
	
	while (o < b->size) {
		n = write(fd, b->data + o, b->size - o);
		
		if (n < 0) {
			if (errno == EINTR) {
				errno = 0;
			} else {
				fprintf(stderr,
					"buffer_write(): write(): %s\n",
					strerror(errno));
				return(-1);
			}
		} else if (n == 0) {
			fprintf(stderr, "buffer_write(): write(): EOF\n");
			return(-1);
		} else {
			o += n;
			b->total_write += n;
		}
	}
	
	return(0);
}

// **************************************************************************
// 
// **************************************************************************
int buffer_load_path(buffer *b, const char *path, size_t max_size)
{
	int fd, r;
	
	fd = open(path, O_RDONLY);
	if (fd < 0) {
		fprintf(stderr, "%s: %s (opening file)\n",
			path, strerror(errno));
		return(-1);
	}
	
	r = buffer_load(b, fd, max_size);
	
	close(fd);
	
	return(r);
}

int buffer_save_path(buffer *b, const char *path)
{
	int fd, r;
	
	fd = creat(path, 0600);
	if (fd < 0) {
		fprintf(stderr, "%s: %s (creating file)\n",
			path, strerror(errno));
		return(-1);
	}
	
	r = buffer_save(b, fd);
	
	close(fd);
	
	return(r);
}

// **************************************************************************
// 
// **************************************************************************
void buffer_io_reset(buffer *b)
{
	b->partial = 0;
	b->eof = 0;
	b->error = 0;
}

int buffer_io_read(buffer *b, int fd)
{
	ssize_t n;
	
	if (b->partial >= b->size)
		return(0);
	
	n = read(fd, b->data + b->partial, b->size - b->partial);
	
	if (n > 0) {
		b->partial += n;
		b->total_read += n;
		return(0);
	}
	
	if (n == 0) {
		b->eof = 1;
		return(-1);
	}
	
	if ((errno == EINTR) || (errno == EAGAIN)) {
		errno = 0;
		return(0);
	}
	
	fprintf(stderr, "buffer_io_read(): read(): %s\n", strerror(errno));
	b->error = 1;
	return(-1);
}

int buffer_io_write(buffer *b, int fd)
{
	ssize_t n;
	
	if (b->partial >= b->size)
		return(0);
	
	n = write(fd, b->data + b->partial, b->size - b->partial);
	
	if (n > 0) {
		b->partial += n;
		b->total_write += n;
		return(0);
	}
	
	if (n == 0) {
		fprintf(stderr, "buffer_write(): write(): EOF\n");
		b->error = 1;
		return(-1);
	}

	if ((errno == EINTR) || (errno == EAGAIN)) {
		errno = 0;
		return(0);
	}
	
	fprintf(stderr, "buffer_write(): write(): %s\n", strerror(errno));
	b->error = 1;
	return(-1);
}

int buffer_io_complete(buffer *b)
{
	return(b->partial >= b->size);
}

int buffer_io_eof(buffer *b)
{
	return(b->eof);
}

int buffer_io_error(buffer *b)
{
	return(b->error);
}

