/*
    This file is part of ben -- a task scheduler for distributed computations
    Copyright (C) 2012-2019  Laurent Poirrier

    ben is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <poll.h>
#include <stdio.h>
#include <limits.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include "xmem.h"
#include "network.h"
#include "block.h"
#include "date.h"
#include "protocol.h"
#include "common.h"
#include "config.h"
#include "buffer.h"
#include "exec.h"
#include "node.h"
#include "control.h"

// **************************************************************************
// 
// **************************************************************************
#define DEBUG(...)	fprintf(stderr, __VA_ARGS__)

#define CONTROL_SEND_ONLY	1

typedef struct control_t {
	const char *action;
	int fd;
	int flags;
	buffer inbuff;
	buffer outbuff;
} control;

// **************************************************************************
// 
// **************************************************************************
static int control_init(control *k, char *action,
	char *addr)
{
	char *name;
	
	k->action = action;

	k->fd = socket_connect(addr);
	
	if (k->fd < 0)
		return(-1);
	
	buffer_init(&k->inbuff, 0);
	buffer_init(&k->outbuff, 0);

	name = xprintf(NULL, "%s: %s", conf.host, action);
	
	proto_append(&k->outbuff, PROTO_CL_MAGIC, PROTO_CL_SETUP,
		"is", -1, name);

	xfree(name);
	
	return(0);
}

static void control_clear(control *k)
{
	close(k->fd);
	buffer_clear(&k->inbuff);
	buffer_clear(&k->outbuff);
}

// **************************************************************************
// 
// **************************************************************************
static void control_print(int h, int tid, int cid, char *host,
	char *d, size_t sz)
{
	FILE *f;
	size_t o, z;
	
	f = (h == PROTO_SV_OUT) ? stdout : stderr;
	
	o = 0;
	while (o < sz) {
		z = o;
		while ((z < sz) && (d[z] != '\n'))
			z++;
		
		fprintf(f, "[%d](%d %s): ", tid, cid, host);
		fwrite(d + o, z - o, 1, f);
		fprintf(f, "%s\n", (z < sz) ? "" : "[...]");
		
		o = z + 1;
	}
}

// **************************************************************************
// 
// **************************************************************************
enum tag_t {
	TAG_JOB_HEADER_BEGIN,
	TAG_JOB_HEADER_END,
	TAG_JOB_ROW_BEGIN,
	TAG_JOB_ROW_END,
	TAG_JOB_EXEC_BEGIN,
	TAG_JOB_EXEC_END,
	TAG_JOB_REGULAR_BEGIN,
	TAG_JOB_REGULAR_END,
	TAG_JOB_DONE_BEGIN,
	TAG_JOB_DONE_END,
	TAG_JOB_RUNNING_BEGIN,
	TAG_JOB_RUNNING_END,
	TAG_JOB_PENDING_BEGIN,
	TAG_JOB_PENDING_END,
	TAG_NODE_HEADER_BEGIN,
	TAG_NODE_HEADER_END,
	TAG_NODE_ROW_BEGIN,
	TAG_NODE_ROW_END,
	TAG_NODE_CONTROL_BEGIN,
	TAG_NODE_CONTROL_END,
	TAG_NODE_COMPUTE_BEGIN,
	TAG_NODE_COMPUTE_END,
	TAG_NODE_IDLE_BEGIN,
	TAG_NODE_IDLE_END,
	TAG_NODE_LOW_BEGIN,
	TAG_NODE_LOW_END,
	TAG_NODE_HIGH_BEGIN,
	TAG_NODE_HIGH_END,
	TAG_NODE_FULL_BEGIN,
	TAG_NODE_FULL_END
};

static char *tag(enum tag_t type)
{
	if (conf.color == BEN_COLOR_HTML) {
		switch (type) {
		case TAG_JOB_HEADER_BEGIN:
			return("<span class='ben-job-header'>");
		case TAG_JOB_HEADER_END:
			return("</span>\n");
		case TAG_JOB_ROW_BEGIN:
			return("");
		case TAG_JOB_ROW_END:
			return("\n");
		case TAG_JOB_EXEC_BEGIN:
			return("<span class='ben-job-exec'>");
		case TAG_JOB_REGULAR_BEGIN:
			return("<span class='ben-job-regular'>");
		case TAG_JOB_DONE_BEGIN:
			return("<span class='ben-job-done'>");
		case TAG_JOB_RUNNING_BEGIN:
			return("<span class='ben-job-running'>");
		case TAG_JOB_PENDING_BEGIN:
			return("<span class='ben-job-pending'>");
		case TAG_NODE_HEADER_BEGIN:
			return("<span class='ben-node-header'>");
		case TAG_NODE_HEADER_END:
			return("</span>\n");
		case TAG_NODE_ROW_BEGIN:
			return("");
		case TAG_NODE_ROW_END:
			return("\n");
		case TAG_NODE_CONTROL_BEGIN:
			return("<span class='ben-node-control'>");
		case TAG_NODE_COMPUTE_BEGIN:
			return("<span class='ben-node-compute'>");
		case TAG_NODE_IDLE_BEGIN:
			return("<span class='ben-node-idle'>");
		case TAG_NODE_LOW_BEGIN:
			return("<span class='ben-node-low'>");
		case TAG_NODE_HIGH_BEGIN:
			return("<span class='ben-node-high'>");
		case TAG_NODE_FULL_BEGIN:
			return("<span class='ben-node-full'>");
		case TAG_JOB_EXEC_END:
		case TAG_JOB_REGULAR_END:
		case TAG_JOB_DONE_END:
		case TAG_JOB_RUNNING_END:
		case TAG_JOB_PENDING_END:
		case TAG_NODE_CONTROL_END:
		case TAG_NODE_COMPUTE_END:
		case TAG_NODE_IDLE_END:
		case TAG_NODE_LOW_END:
		case TAG_NODE_HIGH_END:
		case TAG_NODE_FULL_END:
			return("</span>");
		}
	}
	
	if (conf.color == BEN_COLOR_TTY) {
		switch (type) {
		case TAG_JOB_HEADER_BEGIN:
			return("\033[1m");
		case TAG_JOB_HEADER_END:
			return("\033[0m\n");
		case TAG_JOB_ROW_BEGIN:
			return("");
		case TAG_JOB_ROW_END:
			return("\n");
		case TAG_JOB_EXEC_BEGIN:
			return("\033[36m");
		case TAG_JOB_REGULAR_BEGIN:
		case TAG_JOB_REGULAR_END:
			return("");
		case TAG_JOB_DONE_BEGIN:
			return("\033[34m");
		case TAG_JOB_RUNNING_BEGIN:
			return("\033[1m\033[32m");
		case TAG_JOB_PENDING_BEGIN:
		case TAG_JOB_PENDING_END:
			return("");
		case TAG_NODE_HEADER_BEGIN:
			return("\033[1m");
		case TAG_NODE_HEADER_END:
			return("\033[0m\n");
		case TAG_NODE_ROW_BEGIN:
			return("");
		case TAG_NODE_ROW_END:
			return("\n");
		case TAG_NODE_CONTROL_BEGIN:
			return("\033[2m\033[34m");
		case TAG_NODE_COMPUTE_BEGIN:
		case TAG_NODE_COMPUTE_END:
			return("");
		case TAG_NODE_IDLE_BEGIN:
			return("\033[32m");
		case TAG_NODE_LOW_BEGIN:
			return("\033[36m");
		case TAG_NODE_HIGH_BEGIN:
			return("\033[33m");
		case TAG_NODE_FULL_BEGIN:
			return("\033[31m");
		case TAG_JOB_EXEC_END:
		case TAG_JOB_DONE_END:
		case TAG_JOB_RUNNING_END:
		case TAG_NODE_CONTROL_END:
		case TAG_NODE_IDLE_END:
		case TAG_NODE_LOW_END:
		case TAG_NODE_HIGH_END:
		case TAG_NODE_FULL_END:
			return("\033[0m");
		}
	}
	
	switch (type) {
	case TAG_JOB_HEADER_END:
	case TAG_JOB_ROW_END:
	case TAG_NODE_HEADER_END:
	case TAG_NODE_ROW_END:
		return("\n");
	default:
		return("");
	}
}
static void control_header_job(int flags)
{
	printf("%s", tag(TAG_JOB_HEADER_BEGIN));
	
	printf("#%4s %15s %20s %1s %5s %-14s",
		"id", "dir", "job", "S", "node", "");

	if (flags & BEN_FLAG_DURATIONS) {
		printf("   %8s", "duration");
	}
	
	if (flags & BEN_FLAG_TIMES) {
		printf("   %19s   %19s", "start", "stop");
	}

	printf("%s", tag(TAG_JOB_HEADER_END));
}

static int control_print_job(control *k, int flags)
{
	int tid, brestr, bid, type;
	char *tdir, *tname, *bname;
	uint64_t start, stop, dur;
	char dstart[20], dstop[20];
	int dh, dm, ds;
	
	if (proto_fetch(&k->inbuff, "issiiispp", &tid, &tdir, &tname, &type,
			&brestr, &bid, &bname, &start, &stop)) {
		DEBUG("%s: PROTO_SV_JOB: bad packet\n", k->action);
		return(-1);
	}
	
	printf("%s", tag(TAG_JOB_ROW_BEGIN));
	
	printf("%5d %15s %s%20s%s ", tid, tdir,
		tag((brestr >= 0) ? TAG_JOB_EXEC_BEGIN : TAG_JOB_REGULAR_BEGIN),
		tname,
		tag((brestr >= 0) ? TAG_JOB_EXEC_END : TAG_JOB_REGULAR_END));
	
	if (type == PROTO_TASK_DONE)
		printf("%s.%s",
			tag(TAG_JOB_DONE_BEGIN), tag(TAG_JOB_DONE_END));
	else if (type == PROTO_TASK_RUNNING)
		printf("%sr%s",
			tag(TAG_JOB_RUNNING_BEGIN), tag(TAG_JOB_RUNNING_END));
	else
		printf("%s %s",
			tag(TAG_JOB_PENDING_BEGIN), tag(TAG_JOB_PENDING_END));
	
	if (bid == -1)
		printf(" %5s %-14s", "", bname);
	else
		printf(" %5d %-14s", bid, bname);
	
	if (flags & BEN_FLAG_DURATIONS) {
		dur = stop - start;
		dm = dur / 60;
		ds = dur % 60;
		dh = dm / 60;
		dm = dm % 60;

		if (start)
			printf("  %s%02d:%02d:%02d",
				(type == PROTO_TASK_RUNNING) ? ">" : " ",
				dh, dm, ds);
		else
			printf("   %8s", "");
	}
	
	if (flags & BEN_FLAG_TIMES) {
		if (start)
			date_str(dstart, sizeof(dstart), start);
		else
			dstart[0] = 0;

		if (stop)
			date_str(dstop, sizeof(dstop), stop);
		else
			dstop[0] = 0;
		
		printf("   %19s   %19s", dstart, dstop);
	}
	
	printf("%s", tag(TAG_JOB_ROW_END));
	
	return(0);
}

// **************************************************************************
// 
// **************************************************************************
static void control_header_node(int flags)
{
	printf("%s", tag(TAG_NODE_HEADER_BEGIN));
	
	printf("#%5s %-25s  R  P", "node", "");
	
	if (flags & 0) {
		printf("   in  out ref flags");
	}
	
	printf("%s", tag(TAG_NODE_HEADER_END));
}

static int control_print_node(control *k, int flags)
{
	int bid, run, n, refs, bflags, load0, load1;
	char *bname;
	uint64_t inbsz, outbsz;
	
	if (proto_fetch(&k->inbuff, "isiippii", &bid, &bname,
			&run, &n, &inbsz, &outbsz, &refs, &bflags)) {
		DEBUG("%s: PROTO_SV_NODE: bad packet\n", k->action);
		return(-1);
	}
	
	printf("%s", tag(TAG_NODE_ROW_BEGIN));

	if ((n < 0) && (run == 0)) {
		printf("%s %5d %-25s %2s %2s%s",
			tag(TAG_NODE_CONTROL_BEGIN),
			bid, bname, "-", "C",
			tag(TAG_NODE_CONTROL_END));
	} else {
		if (run == n) {
			load0 = TAG_NODE_FULL_BEGIN;
			load1 = TAG_NODE_FULL_END;
		} else if (run == 0) {
			load0 = TAG_NODE_IDLE_BEGIN;
			load1 = TAG_NODE_IDLE_END;
		} else if (run <= n / 2) {
			load0 = TAG_NODE_LOW_BEGIN;
			load1 = TAG_NODE_LOW_END;
		} else {
			load0 = TAG_NODE_HIGH_BEGIN;
			load1 = TAG_NODE_HIGH_END;
		}
		
		printf(" %5d %-25s %s%2d%s %2d", bid, bname,
			tag(load0), run, tag(load1), n);
	}
	
	// no option for now
	if (flags & 0) {
		printf(" %4ld %4ld %3d %c%c%c\n",
			(long)inbsz, (long)outbsz, refs,
			(bflags & NODE_KICK) ? 'k' : '-',
			(bflags & NODE_SYNC) ? 'S' : '-',
			(bflags & NODE_RETIRE) ? 'R' : '-');
	}
	
	printf("%s", tag(TAG_NODE_ROW_END));
	
	return(0);
}

// **************************************************************************
// 
// **************************************************************************
static int control_iter(control *k, int flags)
{
	void *d;
	size_t dsz;
	uint64_t ping_id;
	int tid, cid;
	char *hs, *s;
	int h;
	
	if (proto_prepare(&k->inbuff, PROTO_SV_MAGIC, conf.proto_buff_max)) {
		DEBUG("%s: protocol error\n", k->action);
		return(-1);
	}
	
	if (!proto_ready(&k->inbuff))
		return(0);
	
	h = proto_header(&k->inbuff);

	// ************************************	
	if (h == PROTO_SV_MSG) {
		if (proto_fetch(&k->inbuff, "d", &d, &dsz)) {
			DEBUG("%s: PROTO_SV_MSG: bad packet\n", k->action);
			return(-1);
		}

		if ((dsz != 0) && (fwrite(d, dsz, 1, stdout) != 1)) {
			DEBUG("fwrite(%zd): %s\n", dsz, strerror(errno));
			return(-1);
		}
		
		return(0);
	}

	// ************************************	
	if ((h == PROTO_SV_OUT) || (h == PROTO_SV_ERR)) {
		hs = (h == PROTO_SV_OUT) ? "OUT" : "ERR";
		
		if (proto_fetch(&k->inbuff, "iisd", &tid, &cid, &s, &d, &dsz)) {
			DEBUG("%s: PROTO_SV_%s: bad packet\n", k->action, hs);
			return(-1);
		}
		
		control_print(h, tid, cid, s, (char *)d, dsz);
		
		return(0);
	}

	// ************************************	
	if (h == PROTO_SV_PING) {
		if (proto_fetch(&k->inbuff, "p", &ping_id)) {
			DEBUG("%s: PROTO_SV_PING: bad packet\n", k->action);
			return(-1);
		}

		proto_append(&k->outbuff, PROTO_CL_MAGIC, PROTO_CL_PONG,
			"p", ping_id);
		
		//DEBUG("[ping] %zd\n", ping_id);
		return(0);
	}

	// ************************************	
	if (h == PROTO_SV_JOB) {
		return(control_print_job(k, flags));
	}

	// ************************************	
	if (h == PROTO_SV_NODE) {
		return(control_print_node(k, flags));
	}

	// ************************************	
	DEBUG("%s: unknown response header 0x%08x\n",
		k->action, h);
	return(-1);
}

// **************************************************************************
// 
// **************************************************************************
static int control_loop(control *k, int flags)
{
	struct pollfd pfd;
	int done, error, detached;
	
	done = 0;
	error = 0;
	detached = 0;
	
	while (1) {
		pfd.fd = k->fd;
		pfd.events = POLLIN;
		if (buffer_io_complete(&k->outbuff)) {
			if ((!detached) && (flags & BEN_FLAG_DAEMONIZE)) {
				if (exec_daemonize(NULL, NULL, NULL))
					return(1);
				detached = 1;
			}
		} else {
			pfd.events |= POLLOUT;
		}
		pfd.revents = 0;
	
		if (poll(&pfd, 1, -1) < 0) {
			fprintf(stderr, "%s: poll(): %s\n",
				k->action, strerror(errno));
			return(1);
		}
		
		if (pfd.revents & (POLLIN|POLLERR|POLLHUP)) {
			buffer_io_read(&k->inbuff, k->fd);
			done |= k->inbuff.eof;
			error |= k->inbuff.error;
		}
		
		if (pfd.revents & (POLLOUT|POLLERR|POLLHUP)) {
			buffer_io_write(&k->outbuff, k->fd);
			error |= k->outbuff.eof | k->outbuff.error;
		}
		
		if ((done) || (error)) {
			error |= (k->inbuff.partial != 0);
			error |= !buffer_io_complete(&k->outbuff);

			control_clear(k);
			
			if (error) {
				fprintf(stderr, "%s: i/o error\n", k->action);
				return(1);
			} else {
				// fprintf(stderr, "-- ok --\n");
				return(0);
			}
		}
		
		if (control_iter(k, flags))
			return(1);
	}
}

// **************************************************************************
// 
// **************************************************************************
int ben_add(char *addr, char *dir, char *cmd, char *opts,
	int n, char **names)
{
	control k;
	int i;
	
	if (control_init(&k, "ben add", addr))
		return(1);
	
	if (!dir)
		dir = "";
	
	if (!cmd)
		cmd = "";
	
	if (!opts)
		opts = "";
	
	for (i = 0; i < n; i++) {
		proto_append(&k.outbuff, PROTO_CL_MAGIC, PROTO_CL_ADD,
			"sssss", conf.shell, dir, opts, names[i], cmd);
	}
	
	proto_append(&k.outbuff, PROTO_CL_MAGIC, PROTO_CL_REQACK, "");
	
	return(control_loop(&k, 0));
}

// **************************************************************************
// 
// **************************************************************************
int ben_rm(char *addr, int mask, char *dir, int n, char **names)
{
	control k;
	char *defz;
	int i;
	
	if (control_init(&k, "ben stop", addr))
		return(1);
	
	if (!dir)
		dir = "";

	if (n == 0) {
		defz = "";
		names = &defz;
		n = 1;
	}
	
	for (i = 0; i < n; i++) {
		proto_append(&k.outbuff, PROTO_CL_MAGIC, PROTO_CL_RM,
			"iss", mask, dir, names[i]);
	}
	
	proto_append(&k.outbuff, PROTO_CL_MAGIC, PROTO_CL_REQACK, "");

	return(control_loop(&k, 0));
}

// **************************************************************************
// 
// **************************************************************************
int ben_exec(char *addr, char *cmd, char *opts, int flags)
{
	control k;
	
	if (control_init(&k, "ben exec", addr))
		return(1);
	
	if (!cmd)
		cmd = "";
	
	if (!opts)
		opts = "";
	
	proto_append(&k.outbuff, PROTO_CL_MAGIC, PROTO_CL_EXEC,
		"sss", conf.shell, opts, cmd);
	
	return(control_loop(&k, flags));
}

// **************************************************************************
// 
// **************************************************************************
int ben_scale(char *addr, int n, int nb, char **b, int flags)
{
	control k;
	int i;
	
	if (control_init(&k, "ben scale", addr))
		return(1);
	
	flags &= PROTO_SCALE_KILL | PROTO_SCALE_RETIRE;
	
	if (nb == 0) {
		proto_append(&k.outbuff, PROTO_CL_MAGIC, PROTO_CL_SCALE,
			"sii", "", n, flags);
	}
	
	for (i = 0; i < nb; i++) {
		proto_append(&k.outbuff, PROTO_CL_MAGIC, PROTO_CL_SCALE,
			"sii", b[i], n, flags);
	}
	
	proto_append(&k.outbuff, PROTO_CL_MAGIC, PROTO_CL_REQACK, "");

	return(control_loop(&k, 0));
}

// **************************************************************************
// 
// **************************************************************************
int ben_list(char *addr, int flags)
{
	control k;
	exec_pager pager;
	int r;
	
	if (control_init(&k, "ben list", addr))
		return(1);
	
	proto_append(&k.outbuff, PROTO_CL_MAGIC, PROTO_CL_LIST, "i", flags);
	
	exec_pager_run(&pager);
	
	control_header_job(flags);
	
	r = control_loop(&k, flags);
	
	exec_pager_wait(&pager);
	
	return(r);
}

// **************************************************************************
// 
// **************************************************************************
int ben_nodes(char *addr)
{
	control k;
	exec_pager pager;
	int r;
	
	if (control_init(&k, "ben nodes", addr))
		return(1);
	
	proto_append(&k.outbuff, PROTO_CL_MAGIC, PROTO_CL_NODES, "");
	
	exec_pager_run(&pager);
	
	control_header_node(0);

	r = control_loop(&k, 0);

	exec_pager_wait(&pager);

	return(r);
}


// **************************************************************************
// 
// **************************************************************************
int ben_exit(char *addr, int flags)
{
	control k;
	
	if (control_init(&k, "ben exit", addr))
		return(1);
	
	flags &= PROTO_SCALE_RETIRE;
	
	proto_append(&k.outbuff, PROTO_CL_MAGIC, PROTO_CL_EXIT, "i", flags);
	
	return(control_loop(&k, 0));
}
