/*
    This file is part of ben -- a task scheduler for distributed computations
    Copyright (C) 2012-2019  Laurent Poirrier

    ben is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <string.h>
#include <stdlib.h>
#include "buffer.h"
#include "strlist.h"
#include "config.h"
#include "param.h"


// ***************************************************************************
// 
// ***************************************************************************
static void param_format(FILE *f, param_desc *d)
{
	if (d->type == PARAM_NEWLINE) {
		fprintf(f, "\n   ");
		return;
	}

	fprintf(f, (d->minocc == 0) ? "[" : "{");
		
	if (d->load) {
		if (d->load[0] == 0)
			fprintf(f, "%s", d->loadname);
		else
			fprintf(f, "%s %s", d->load, d->loadname);

		if (d->imm)
			fprintf(f, "|");
	}
	
	if (d->imm) {
		if (d->imm[0] == 0)
			fprintf(f, "%s", d->immname);
		else if ((!d->immname) || (d->immname[0] == 0))
			fprintf(f, "%s", d->imm);
		else
			fprintf(f, "%s %s", d->imm, d->immname);
		
		if (d->type == PARAM_STRLIST)
			fprintf(f, "...");
	}

	fprintf(f, (d->minocc == 0) ? "]" : "}");
}

// ***************************************************************************
// 
// ***************************************************************************
void param_help(FILE *f, param_set *ps)
{
	int i;
	
	fprintf(f, "%s", ps->exec);
	
	for (i = 0; i < ps->n; i++) {
		fprintf(f, " ");
		param_format(f, &ps->d[i]);
	}
	
	fprintf(f, "\n");
}



// ***************************************************************************
// 
// ***************************************************************************
static int param_match_strict(const char *s, const char *template)
{
	if (!template)
		return(0);
	
	if (template[0] == 0)
		return(0);
	
	return(!strcmp(s, template));
}

static int param_match_all(const char *template)
{
	if (!template)
		return(0);
	
	if (template[0] == 0)
		return(1);
	
	return(0);
}

// ***************************************************************************
// 
// ***************************************************************************
static int param_do(param_desc *d, param_data *p,
	int ld, const char *exec, int na, char **a, int t)
{
	char *s, *e;
	int r;
	
	p->occ++;
	
	if ((d->type == PARAM_STRLIST) && (!ld))
		p->occ = 1;
	
	if ((d->maxocc >= 0) && (p->occ > d->maxocc)) {
		fprintf(stderr, "%s: %s: parameter ", exec, a[t]);
		param_format(stderr, d);
		fprintf(stderr, " specified %d times\n", p->occ);
		return(-1);
	}
	
	if (d->type == PARAM_SWITCH) {
		p->value.b = 1;
		return(t + 1);
	}
	
	if (((!ld) && (d->imm[0] == 0)) || ((ld) && (d->load[0] == 0))) {
		s = a[t];
		r = t + 1;
	} else {
		if (t + 1 >= na) {
			fprintf(stderr, "%s: Expected: %s %s\n",
				exec, a[t], (ld) ? d->loadname : d->immname);
			return(-1);
		}
		
		s = a[t + 1];
		r = t + 2;
	}
	
	if (ld) {
		if (buffer_load_path(&p->buff, s, conf.file_size_max))
			return(-1);
		
		s = buffer_str(&p->buff);
	}
	
	if (d->type == PARAM_STRING) {
		p->value.s = s;
	} else if (d->type == PARAM_INT) {
		p->value.i = strtol(s, &e, 0);
		if ((*e != 0) || (e == s)) {
			fprintf(stderr, "%s: %s: invalid integer '%s'\n",
				exec, a[t], s);
			return(-1);
		}
	} else if (d->type == PARAM_DOUBLE) {
		p->value.d = strtod(s, &e);
		if ((*e != 0) || (e == s)) {
			fprintf(stderr, "%s: %s: invalid number '%s'\n",
				exec, a[t], s);
			return(-1);
		}
	} else if (d->type == PARAM_STRLIST) {
		if (!ld)
			strlist_append_one(&p->value.l, s);
		else
			strlist_parse(&p->value.l, buffer_str(&p->buff));
	}
	
	return(r);
}

// ***************************************************************************
// 
// ***************************************************************************
int param_init(param_set *ps, int na, char **a)
{
	param_desc *d = ps->d;
	param_data *p = ps->p;
	int np = ps->n;
	char *exec = ps->exec;
	int i, t;
	
	// Init
	for (i = 0; i < np; i++) {
		buffer_init(&p[i].buff, conf.block_size);
		if (d[i].type == PARAM_STRING)
			p[i].value.s = NULL;
		else if (d[i].type == PARAM_INT)
			p[i].value.i = -1;
		else if (d[i].type == PARAM_DOUBLE)
			p[i].value.d = 0.0;
		else if (d[i].type == PARAM_STRLIST)
			strlist_init(&p[i].value.l);
		else if (d[i].type == PARAM_SWITCH)
			p[i].value.b = 0;
		else if (d[i].type == PARAM_NEWLINE)
			p[i].value.b = 0;

		p[i].occ = 0;
	}

	// Parse
	t = 0;
	while ((t >= 0) && (t < na)) {
		for (i = 0; i < np; i++)
		if ((t >= d[i].minpos)
		&&  ((d[i].maxpos < 0) || (t <= d[i].maxpos))) {
			// one of imm and load could be empty,
			// it then has to be checked *after* the other one
			if (param_match_strict(a[t], d[i].imm)) {
				t = param_do(&d[i], &p[i], 0, exec, na, a, t);
				break;
			}
			if (param_match_strict(a[t], d[i].load)) {
				t = param_do(&d[i], &p[i], 1, exec, na, a, t);
				break;
			}
			if (param_match_all(d[i].imm)) {
				t = param_do(&d[i], &p[i], 0, exec, na, a, t);
				break;
			}
			if (param_match_all(d[i].load)) {
				t = param_do(&d[i], &p[i], 1, exec, na, a, t);
				break;
			}
		}
		
		if (i == np) {
			fprintf(stderr, "%s: unknown parameter %s\n",
				exec, a[t]);
			t = -1;
		}
	}
	
	if (t < 0) {
		param_clear(ps);
		return(-1);
	}
	
	// Check
	for (i = 0; i < np; i++) {
		if (p[i].occ < d[i].minocc) {
			fprintf(stderr, "%s: parameter ", exec);
			param_format(stderr, &d[i]);
			fprintf(stderr, " needed\n");
			t = -1;
		}
	}
	
	if (t < 0) {
		param_clear(ps);
		return(-1);
	}
	
	return(0);
}

// ***************************************************************************
// 
// ***************************************************************************
void param_clear(param_set *ps)
{
	param_desc *d = ps->d;
	param_data *p = ps->p;
	int np = ps->n;
	int i;
	
	for (i = 0; i < np; i++) {
		buffer_clear(&p[i].buff);
		if (d[i].type == PARAM_STRLIST)
			strlist_clear(&p[i].value.l);
	}
}


