/*
    This file is part of ben -- a task scheduler for distributed computations
    Copyright (C) 2012-2019  Laurent Poirrier

    ben is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdint.h>
#include "protocol.h"

// **************************************************************************
// Debug
// **************************************************************************
#define DEBUG(...)	fprintf(stderr, __VA_ARGS__)

// **************************************************************************
// 
// **************************************************************************
static void copy_int(void *p, int v)
{
	unsigned char *byte;
	
	byte = (unsigned char *)p;
	byte[0] = v & 0xff;
	byte[1] = (v >> 8) & 0xff;
	byte[2] = (v >> 16) & 0xff;
	byte[3] = (v >> 24) & 0xff;
}

static void copy_u64(void *p, uint64_t v)
{
	unsigned char *byte;

	byte = (unsigned char *)p;	
	byte[0] = v & 0xff;
	byte[1] = (v >> 8) & 0xff;
	byte[2] = (v >> 16) & 0xff;
	byte[3] = (v >> 24) & 0xff;
	byte[4] = (v >> 32) & 0xff;
	byte[5] = (v >> 40) & 0xff;
	byte[6] = (v >> 48) & 0xff;
	byte[7] = (v >> 56) & 0xff;
}

static int fetch_int(void *p)
{
	unsigned char *byte;
	int v;
	
	byte = (unsigned char *)p;
	v = (int)byte[0] | ((int)byte[1] << 8) | ((int)byte[2] << 16)
		| ((int)byte[3] << 24);
	
	return(v);
}

static uint64_t fetch_u64(void *p)
{
	unsigned char *byte;
	uint64_t v;

	byte = (unsigned char *)p;	
	v =	  ((uint64_t)byte[0])
		| ((uint64_t)byte[1] << 8)
		| ((uint64_t)byte[2] << 16)
		| ((uint64_t)byte[3] << 24)
		| ((uint64_t)byte[4] << 32)
		| ((uint64_t)byte[5] << 40)
		| ((uint64_t)byte[6] << 48)
		| ((uint64_t)byte[7] << 56);
	
	return(v);
}

// **************************************************************************
// Sending internals
// **************************************************************************
static void proto_append_int(buffer *b, int v)
{
	unsigned char byte[4];
	
	copy_int(byte, v);
	buffer_memappend(b, byte, 4);
}

static void proto_append_u64(buffer *b, uint64_t v)
{
	unsigned char byte[8];
	
	copy_u64(byte, v);
	buffer_memappend(b, byte, 8);
}

static void proto_append_string(buffer *b, char *s)
{
	size_t l;
	
	l = strlen(s);
	
	proto_append_u64(b, l + 1);
	buffer_memappend(b, s, l + 1);
}

static void proto_append_data(buffer *b, void *data, size_t sz)
{
	proto_append_u64(b, sz);
	buffer_memappend(b, data, sz);
}

// **************************************************************************
// Sending internals
// **************************************************************************
static int proto_fetch_int(buffer *b, size_t *offset, int *r)
{
	int v;
	size_t o;
	
	o = *offset;
	
	if (o + 4 > b->size) {
		DEBUG("proto_fetch_int(): at %zd / %zd\n", o, b->size);
		return(-1);
	}
	
	v = fetch_int(b->data + o);
	
	if (r)
		*r = v;
	*offset = o + 4;
	return(0);
}

static int proto_fetch_u64(buffer *b, size_t *offset, uint64_t *r)
{
	uint64_t v;
	size_t o;
	
	o = *offset;
	
	if (o + 8 > b->size) {
		DEBUG("proto_fetch_u64(): at %zd / %zd\n", o, b->size);
		return(-1);
	}
	
	v = fetch_u64(b->data + o);
	
	if (r)
		*r = v;
	*offset = o + 8;
	return(0);
}

static int proto_fetch_string(buffer *b, size_t *offset, char **r)
{
	char *v;
	uint64_t sz;
	size_t o;
	
	if (proto_fetch_u64(b, offset, &sz)) {
		DEBUG("proto_fetch_string(): proto_fetch_u64() failed\n");
		return(-1);
	}
	
	o = *offset;
	
	if (o + sz > b->size) {
		DEBUG("proto_fetch_string(): at %zd / %zd, size %zd\n",
			o, b->size, (size_t)sz);
		return(-1);
	}
	
	v = ((char *)b->data) + o;
	
	if (v[sz - 1] != 0) {
		DEBUG("proto_fetch_string(): not zero-terminated "
			"(str[%zd] = %d)\n",
			(size_t)sz - 1, v[sz - 1]);
		return(-1);
	}
	
	if (r)
		*r = v;
		
	*offset = o + sz;
	return(0);
}

static int proto_fetch_data(buffer *b, size_t *offset, void **r, size_t *rsz)
{
	uint64_t sz;
	size_t o;
	
	if (proto_fetch_u64(b, offset, &sz)) {
		DEBUG("proto_fetch_data(): proto_fetch_size() failed\n");
		return(-1);
	}
	
	o = *offset;
	
	if (o + sz > b->size) {
		DEBUG("proto_fetch_data(): at %zd / %zd, size %zd\n",
			o, b->size, (size_t)sz);
		return(-1);
	}

	if (r)
		*r = ((void *)b->data) + o;
	
	if (rsz)
		*rsz = sz;
		
	*offset = o + sz;
	return(0);
}

// **************************************************************************
// Sending
// **************************************************************************
void proto_append(buffer *b, int magic, int header, const char *fmt, ...)
{
	va_list ap;
	const char *f;
	size_t ipos;
	int i;
	uint64_t p;
	char *s;
	void *d;
	
	proto_append_int(b, magic);
	proto_append_int(b, header);
	ipos = b->size;
	proto_append_u64(b, 0);
	
	f = fmt;
	va_start(ap, fmt);

	while (*f) {
		if (*f == 'i') {
			i = va_arg(ap, int);
			proto_append_int(b, i);
		} else if (*f == 'p') {
			p = va_arg(ap, uint64_t);
			proto_append_u64(b, p);
		} else if (*f == 's') {
			s = va_arg(ap, char *);
			proto_append_string(b, s);
		} else if (*f == 'd') {
			d = va_arg(ap, void *);
			p = va_arg(ap, size_t);
			proto_append_data(b, d, p);
		} else {
			DEBUG("proto_append(): unknown format '%c'\n", *f);
			exit(1);
		}
		f++;
	}	
	
	copy_u64(b->data + ipos, b->size - (ipos + 8));
	
	va_end(ap);
}


// **************************************************************************
// Receiving
// **************************************************************************
int proto_prepare(buffer *b, int magic, size_t max_size)
{
	int m;
	size_t sz;

	if (b->size == 0) {
		buffer_get(b, 16);
		return(0);
	}

	if (b->partial < 16)
		return(0);
	
	m = fetch_int(b->data);
	if (m != magic) {
		DEBUG("proto_prepare(): magic number mismatch: "
			"0x%08x != 0x%08x\n", m, magic);
		return(-1);
	}
	
	sz = 16 + fetch_u64(b->data + 8);
	if ((sz < 16) || (sz > max_size)) {
		DEBUG("proto_prepare(): invalid packet size: %zd "
			"(min 16, max %zd)\n", sz, max_size);
		return(-1);
	}

	buffer_get(b, sz);
	return(0);
}

int proto_ready(buffer *b)
{
	return(b->partial == b->size);
}

int proto_header(buffer *b)
{
	if (b->partial < 16) {
		DEBUG("proto_header(): header not full: %zd / %zd, 16 needed\n",
			b->partial, b->size);
		return(-1);
	}
	
	return(fetch_int(b->data + 4));
}

int proto_fetch(buffer *b, const char *fmt, ...)
{
	va_list ap;
	size_t offset;
	int *ip;
	uint64_t *pp;
	size_t *zp;
	char **sp;
	void **dp;
	const char *f;
	int r;
	
	if (b->partial != b->size) {
		DEBUG("proto_fetch(): buffer not full: %zd / %zd\n",
			b->partial, b->size);
		return(-1);
	}
	
	f = fmt;
	r = 0;
	offset = 16;
	
	va_start(ap, fmt);

	while (*f) {
		if (*f == 'i') {
			ip = va_arg(ap, int *);
			r = proto_fetch_int(b, &offset, ip);
		} else if (*f == 'p') {
			pp = va_arg(ap, uint64_t *);
			r = proto_fetch_u64(b, &offset, pp);
		} else if (*f == 's') {
			sp = va_arg(ap, char **);
			r = proto_fetch_string(b, &offset, sp);
		} else if (*f == 'd') {
			dp = va_arg(ap, void **);
			zp = va_arg(ap, size_t *);
			r = proto_fetch_data(b, &offset, dp, zp);
		} else {
			DEBUG("proto_fetch(): unknown format '%c'\n", *f);
			exit(1);
		}
		
		if (r)
			break;
		
		f++;
	}	
	
	va_end(ap);
	
	if (r)
		return(-1);
	
	if (offset != b->size) {
		DEBUG("proto_fetch(): trailing garbage %zd - %zd\n",
			offset, b->size);
		return(-1);
	}
	
	// upon successful packet retrieval, clear the buffer
	buffer_reset(b);
	buffer_io_reset(b);
	
	return(0);
}


