/*
    This file is part of ben -- a task scheduler for distributed computations
    Copyright (C) 2012-2019  Laurent Poirrier

    ben is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <stdio.h>
#include <stdint.h>
#include <poll.h>
#include <errno.h>
#include <string.h>
#include "block.h"
#include "xmem.h"
#include "exec.h"
#include "network.h"
#include "ssh.h"

#define DEBUG(...)	fprintf(stderr, __VA_ARGS__)

// **************************************************************************
//
// **************************************************************************
static int ssh_needs_quote(char *s)
{
	while (*s) {
		if ((*s <= ' ') || (*s >= 127))
			return(1);
		
		s++;
	}
	
	return(0);
}

static void ssh_argv_dump(char *prefix, char **argv)
{
	char *quote;
	int i;
	
	DEBUG("%s", prefix);
	
	i = 0;
	while (argv[i]) {
		quote = (ssh_needs_quote(argv[i])) ? "'" : "";
		DEBUG(" %s%s%s", quote, argv[i], quote);
		i++;
	}
	
	DEBUG("\n");
}

// **************************************************************************
//
// **************************************************************************
int ssh_fwd_fork(exec_unit *u, char *dst, char *laddr, char *faddr)
{
	socket_addr a;
	char *argv[8];
	char *setup;
	uint32_t ping, pong;
	int r;
	
	if (socket_addr_parse(&a, laddr)) {
		DEBUG("ssh_fwd_fork(): could not parse local address '%s'.\n",
			laddr);
		return(-1);
	}
	
	if (a.type == SOCKET_UNIX) {
		if (unlink(a.path))
			errno = 0;
		
		if (exec_path_mkdir(a.path)) {
			DEBUG("ssh_fwd_fork(): warning: could not create"
				" directory for '%s'\n", a.path);
			errno = 0;
		}
	}
	
	setup = xprintf(NULL, "%s:%s%s",
		laddr,
		(faddr[0] == ':') ? "localhost" : "", faddr);
	
	argv[0] = "ssh";
	argv[1] = "-o";
	argv[2] = "ExitOnForwardFailure=yes";
	//argv[] = "-o";
	//argv[] = "StreamLocalBindUnlink=yes";
	argv[3] = "-L";
	argv[4] = setup;
	argv[5] = dst;
	argv[6] = "cat";
	argv[7] = NULL;
	
	r = exec_fork(u, argv, NULL, EXEC_KEEP_STDERR);
	
	if (r) {
		xfree(setup);
		return(-1);
	}
	
	ping = 0x30405060;
	
	if (block_write(u->infd, &ping, sizeof(ping))) {
		exec_terminate(u, NULL);
		DEBUG("ssh_fwd_fork(): write error\n");
		r = -1;
	}
	
	if (block_read(u->outfd, &pong, sizeof(pong))) {
		exec_terminate(u, NULL);
		DEBUG("ssh_fwd_fork(): read error\n");
		r = -1;
	}
	
	if (ping != pong) {
		exec_terminate(u, NULL);
		DEBUG("ssh_fwd_fork(): cat: %lx != %lx\n",
			(long)ping, (long)pong);
		r = -1;
	}
	
	if (r)
		ssh_argv_dump("command:", argv);
	
	xfree(setup);
	return(r);
}


// **************************************************************************
//
// **************************************************************************
int ssh_fwd_terminate(exec_unit *u)
{
	int status;
	
	if (exec_terminate(u, &status))
		return(-1);
	
	if (status)
		return(-1);
	
	return(0);
}

// **************************************************************************
//
// **************************************************************************
#define SSH_POS_MATCH		-1
#define SSH_E_BUFF		128

typedef struct sshr_t {
	exec_unit u;
	int eof;
	char *m_str;
	int m_pos;
	char e_buff[SSH_E_BUFF];
	int e_pos;
} sshr;

static int ssh_remote_iter(sshr *s)
{
	char buff[1024];
	struct pollfd pfd[2];
	int i, t;
	ssize_t n;
	
	if (s->eof == 3)
		return(0);
	
	for (i = 0; i <= 1; i++) {
		pfd[i].fd = (s->eof & (1 << i)) ? -1
			: ((i == 0) ? s->u.outfd : s->u.errfd);
		pfd[i].events = POLLIN;
		pfd[i].revents = 0;
	}
	
	if (poll(pfd, 2, -1) < 0) {
		DEBUG("ssh_remote_exec(): poll(): %s\n",
			strerror(errno));
		return(-1);
	}
	
	for (i = 0; i <= 1; i++) {
		if (pfd[i].revents & (POLLIN|POLLERR|POLLHUP)) {
			n = read(pfd[i].fd, buff, sizeof(buff));
			
			if (n < 0) {
				DEBUG("ssh_remote_exec(): read(): %s\n",
					strerror(errno));
				return(-1);
			}
			
			if (n == 0) {
				s->eof |= 1 << i;
				continue;
			}
			
			if ((i == 0) && (s->m_str)) {
				for (t = 0; t < n; t++) {
					if (buff[t] == s->m_str[s->m_pos])
						s->m_pos++;
					else
						s->m_pos = 0;
					
					if (s->m_str[s->m_pos] == 0) {
						s->m_str = NULL;
						s->m_pos = SSH_POS_MATCH;
						break;
					}
				}
			}
			
			if ((i == 1) && (s->e_pos < SSH_E_BUFF)) {
				for (t = 0; t < n; t++) {
					if (s->e_pos >= SSH_E_BUFF)
						break;

					s->e_buff[s->e_pos] = buff[t];
					s->e_pos++;
				}
			}
			
			if (block_write(1 + i, buff, n)) {
				DEBUG("ssh_remote_exec(): write error\n");
				return(-1);
			}
		}
	}
	
	return(1);
}

// **************************************************************************
//
// **************************************************************************
static int ssh_remote_error_match(sshr *s, char *str)
{
	if (s->e_pos == SSH_E_BUFF)
		s->e_pos--;
	s->e_buff[s->e_pos] = 0;

	return(strstr(s->e_buff, str) != NULL);
}

// **************************************************************************
//
// **************************************************************************
#define SSH_REMOTE_MSG_LISTEN	\
	"Error: remote port forwarding failed for listen path"
#define SSH_REMOTE_MSG_NOTFOUND	\
	"ben: command not found"
#define SSH_REMOTE_ERROR_LISTEN		0x0001
#define SSH_REMOTE_ERROR_NOTFOUND	0x0002
#define SSH_REMOTE_ERROR_OTHER		0x0004
static int ssh_remote_helper(char **argv, char *resp, char *logfile)
{
	sshr s;
	int r, status;
	
	//ssh_argv_dump("command:", argv);
	
	r = exec_fork(&s.u, argv, NULL, 0);
	
	if (r) {
		ssh_argv_dump("command:", argv);
		return(SSH_REMOTE_ERROR_OTHER);
	}
	
	s.eof = 0;
	
	s.m_str = resp;
	s.m_pos = 0;
	memset(s.e_buff, 0, SSH_E_BUFF);
	s.e_pos = 0;
	
	if ((s.m_str) && (s.m_str[0] == 0)) {
		// empty string, immediate match
		s.m_str = NULL;
		s.m_pos = SSH_POS_MATCH;
	}

	do {
		if (s.m_pos == SSH_POS_MATCH) {
			if (exec_daemonize(NULL, NULL, logfile)) {
				r = -1;
				break;
			}
			s.m_pos = 0;
			s.u.flags |= EXEC_ORPHANED;
		}
		
		r = ssh_remote_iter(&s);
	} while (r > 0);
	
	exec_terminate(&s.u, &status);
	
	if ((r) || (status != 0)) {
		ssh_argv_dump("command was:", argv);
	}
	
	if ((r == 0) && (status != 0)) {
		if (ssh_remote_error_match(&s, SSH_REMOTE_MSG_LISTEN))
			return(SSH_REMOTE_ERROR_LISTEN);
		if (ssh_remote_error_match(&s, SSH_REMOTE_MSG_NOTFOUND))
			return(SSH_REMOTE_ERROR_NOTFOUND);
	}
	
	if (r)
		return(SSH_REMOTE_ERROR_OTHER);
	
	return(0);
}

// **************************************************************************
//
// **************************************************************************
int ssh_remote_exec(char *dst, char *raddr, char *laddr,
	char *cmd, char *resp, char *logfile)
{
	char *argv_cmd[8];
	char *argv_mkd[4];
	char *setup, *dir, *slash, *mkd;
	int r, mask;
	
	setup = xprintf(NULL, "%s:%s%s",
		raddr,
		(laddr[0] == ':') ? "localhost" : "", laddr);
	
	argv_cmd[0] = "ssh";
	argv_cmd[1] = "-o";
	argv_cmd[2] = "ExitOnForwardFailure=yes";
	//argv[] = "-o";
	//argv[] = "StreamLocalBindUnlink=yes";
	argv_cmd[3] = "-R";
	argv_cmd[4] = setup;
	argv_cmd[5] = dst;
	argv_cmd[6] = cmd;
	argv_cmd[7] = NULL;
	
	dir = xstrdup(raddr);
	slash = strrchr(dir, '/');
	
	if ((!slash) || (slash == dir)) {
		mkd = NULL;
	} else {
		*slash = 0;
		mkd = xprintf(NULL, "mkdir -p -m 0700 '%s'", dir);
	}
	
	argv_mkd[0] = "ssh";
	argv_mkd[1] = dst;
	argv_mkd[2] = mkd;
	argv_mkd[3] = NULL;
	
	//ssh_argv_dump("command:", argv_cmd);
	//ssh_argv_dump("  mkdir:", argv_mkd);
	
	mask = 0;
	
	while (1) {
		r = ssh_remote_helper(argv_cmd, resp, logfile);
		
		if (((r & (~mask)) == SSH_REMOTE_ERROR_LISTEN) && (mkd)) {
			mask |= SSH_REMOTE_ERROR_LISTEN;
			DEBUG("\n\n"
				"ssh could not create socket '%s' on remote.\n"
				"Attempting to create directory '%s'.\n\n",
				raddr, dir);
		
			r = ssh_remote_helper(argv_mkd, NULL, NULL);
		
			if (r == 0) {
				DEBUG("\n\n"
					"Directory Created. "
					"Retrying ssh forward.\n\n");
				continue;
			}
		}

		if ((r & (~mask)) == SSH_REMOTE_ERROR_NOTFOUND) {
			DEBUG("\n\n"
				"ben executable not found on remote.\n"
				"Try specifying its full path with "
				"--remote-path.\n\n");
		}
		
		break;
	}
	
	xfree(setup);
	xfree(dir);
	xfree(mkd);
	return(r);
}

// **************************************************************************
//
// **************************************************************************
